import os
import re
import string

import qtawesome
from PyQt5.QtCore import QUrl
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QFrame, QLabel, QLineEdit, QCheckBox, \
    QDialogButtonBox, QPlainTextEdit, QHBoxLayout, QToolButton

from dpp.core.exception import AbortedException
from dpp.core.plugin import ScriptPlugin, PluginConfig
from dpp.ui.dialog.plugin_config_dialog import PluginConfigPreviewDialog


class Plugin(ScriptPlugin):
    """
    Reformats the input.

    Example:

        Input:
            123 456

        Parameters:

                    Format = "{1} {0}"
                  Split by = " "
                     Regex = False
            Handle Newline = True

        Output:
            456 123
    """

    class Option(object):

        Format = PluginConfig.Option.Label("format_string", "Format:")
        SplitChars = PluginConfig.Option.Label("split_chars", "Split by:")
        IsRegex = PluginConfig.Option.Label("is_regex", "Regex")
        HandleNewlines = PluginConfig.Option.Label("handle_newlines", "Handle Newlines")

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Reformat Text', "Thomas Engel", [], context)

        def _validate_split_chars(config, codec, input):

            def _validate_regex():
                try:
                    re.compile(config.get(Plugin.Option.SplitChars).value)
                    return True
                except:
                    return False

            if not config.get(Plugin.Option.SplitChars).value:
                return "Split by should not be empty!"

            if config.get(Plugin.Option.IsRegex).value and not _validate_regex():
                return "Invalid regular expression!"

        def _validate_format(config, codec, input):
            try:
                codec(config, input)
            except:
                return "Invalid format string!"

        self.config.add(PluginConfig.Option.String(
            label=Plugin.Option.Format,
            value="",
            description="the format string to be used",
            is_required=True
        ), validator=_validate_format)
        self.config.add(PluginConfig.Option.String(
            label=Plugin.Option.SplitChars,
            value=" ",
            description="the characters used to split the text in individual parts (default=' ')",
            is_required=False
        ), validator=_validate_split_chars)
        self.config.add(PluginConfig.Option.Boolean(
            label=Plugin.Option.IsRegex,
            value=False,
            description="defines whether the split chars is a regular expression (default=False)",
            is_required=False
        ))
        self.config.add(PluginConfig.Option.Boolean(
            label=Plugin.Option.HandleNewlines,
            value=True,
            description="defines whether the operation should be applied for each individual line (default=True)",
            is_required=False
        ))
        self._dialog = None
        self._codec = ReformatCodec()

    def select(self, text: str):
        if not self._dialog:
            self._dialog = PluginConfigPreviewDialog(self._context, self.config.clone(),
                                              "Reformat Text", self._codec.reformat, qtawesome.icon("fa.edit"))

        self._dialog.setInput(text)
        self._dialog_return_code = self._dialog.exec_()

        if self._dialog_return_code != QDialog.Accepted:
            # User clicked the Cancel-Button.
            raise AbortedException("Aborted")

        self.config.update(self._dialog.config)
        return self.run(text)


    def title(self):
        return "Reformat text with '{}' using '{}' as {}delimiter{}".format(
            self.config.get(Plugin.Option.Format).value,
            self.config.get(Plugin.Option.SplitChars).value,
            "regex-" if self.config.get(Plugin.Option.IsRegex).value else "",
            " (newline sensitive)" if self.config.get(Plugin.Option.HandleNewlines).value else ""
        )

    def run(self, text):
        if text:
            return self._codec.reformat(self.config, text)
        return ''


class ReformatCodec:

    class SafeDict(dict):
        def __missing__(self, key):
            return '{' + key + '}'

    def reformat(self, config, input):

        format = config.get(Plugin.Option.Format).value
        split_by_chars = config.get(Plugin.Option.SplitChars).value
        is_regex = config.get(Plugin.Option.IsRegex).value
        handle_newlines = config.get(Plugin.Option.HandleNewlines).value

        def _fill_blanks(format, values):
            """ Ensure that there are always at least as many values as there are placeholders. """
            format_len = len([i for i in string.Formatter().parse(format)])
            if len(values) < format_len:
                for i in range(1, format_len - len(values)):
                    values.append("")
            return values

        def _reformat(text):
            if is_regex:
                split_input = re.split(split_by_chars, text)
            else:
                split_input = text.split(split_by_chars)
            return format.format(*_fill_blanks(format, split_input))

        if input:
            if format:
                try:
                    if handle_newlines:
                        return os.linesep.join([_reformat(line) for line in input.split(os.linesep)])
                    else:
                        return _reformat(input)
                except BaseException:
                    raise Exception("Error during reformat operation! Check your format string!")
            else:
                return input
        return ''


class ReformatDialog(QDialog):

    def __init__(self, context, input, config, codec):
        super(ReformatDialog, self).__init__()
        self._context = context
        self._input = input
        self.config = config
        self._codec = codec
        self.setLayout(self._init_main_layout())
        self._update_view()
        self.setWindowIcon(qtawesome.icon("fa.edit"))
        self.setWindowTitle("Reformat Text")

    def _init_main_layout(self):
        frm_main_layout = QVBoxLayout()
        frm_main_layout.addWidget(self._init_form_frame())
        self._txt_preview = QPlainTextEdit(self)
        self._txt_preview.setReadOnly(True)
        self._txt_preview.setFixedHeight(126)
        frm_main_layout.addWidget(self._txt_preview)
        frm_main_layout.addWidget(self._init_error_frame())
        self._btn_box = self._init_button_box()
        frm_main_layout.addWidget(self._btn_box)
        return frm_main_layout

    def _init_error_frame(self):
        self._error_frame = QFrame()
        layout = QVBoxLayout()
        self._error_text = QLabel("")
        self._error_text.setStyleSheet('QLabel { color: red }')
        layout.addWidget(self._error_text)
        self._error_frame.setLayout(layout)
        self._error_frame.setHidden(True)
        return self._error_frame

    def _init_form_frame(self):
        # Define input fields and register change events
        self._txt_format = QLineEdit()
        self._txt_format.setText(self.config.get(Plugin.Option.Format).value)
        self._txt_format.textChanged.connect(self._on_change_event)

        self._btn_format_help = QToolButton()
        self._btn_format_help.setText("?")
        self._btn_format_help.clicked.connect(lambda evt: QDesktopServices.openUrl(QUrl("https://pyformat.info/")))

        self._txt_split_char = QLineEdit()
        self._txt_split_char.setText(self.config.get(Plugin.Option.SplitChars).value)
        self._txt_split_char.textChanged.connect(self._on_change_event)

        self._chk_regex = QCheckBox("Regex")
        self._chk_regex.setChecked(self.config.get(Plugin.Option.IsRegex).value)
        self._chk_regex.clicked.connect(self._on_change_event)

        self._chk_new_lines = QCheckBox("Handle Newline")
        self._chk_new_lines.setChecked(self.config.get(Plugin.Option.HandleNewlines).value)
        self._chk_new_lines.clicked.connect(self._on_change_event)

        # Build layout
        frm_frame = QFrame()
        frm_layout = QHBoxLayout()
        frm_layout.addWidget(QLabel("Format: "))
        frm_layout.addWidget(self._txt_format)
        frm_layout.addWidget(self._btn_format_help)
        frm_layout.addWidget(QLabel("Split by: "))
        frm_layout.addWidget(self._txt_split_char)
        frm_layout.addWidget(self._chk_regex)
        frm_layout.addWidget(self._chk_new_lines)
        frm_frame.setLayout(frm_layout)
        return frm_frame

    def _init_button_box(self):
        btn_box = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        btn_box.accepted.connect(self.accept)
        btn_box.rejected.connect(self.reject)
        return btn_box

    def _validate_regex(self):
        try:
            if self._is_regex():
                re.compile(self._get_format_text())
            return True
        except:
            return False

    def _update_view(self):
        self._txt_format.setText(self.config.get(Plugin.Option.Format).value)
        self._txt_split_char.setText(self.config.get(Plugin.Option.SplitChars).value)
        self._chk_regex.setChecked(self.config.get(Plugin.Option.IsRegex).value)
        self._chk_new_lines.setChecked(self.config.get(Plugin.Option.HandleNewlines).value)
        self._on_change_event(None)

    def _on_change_event(self, event):
        if not self._txt_split_char.text():
            self._show_error_message(self._txt_split_char, "Split by text should not be empty!")
            self._btn_box.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        if self._chk_regex.isChecked() and self._validate_regex():
            self._show_error_message(self._txt_split_char, "Invalid regular expression!")
            self._btn_box.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        if not self._do_preview():
            self._show_error_message(self._txt_format, "Invalid format string!")
            self._btn_box.button(QDialogButtonBox.Ok).setEnabled(False)
            return

        self._hide_error_message([self._txt_format, self._txt_split_char])
        self._btn_box.button(QDialogButtonBox.Ok).setEnabled(True)

        self.config.get(Plugin.Option.Format).value = self._txt_format.text()
        self.config.get(Plugin.Option.IsRegex).value = self._chk_regex.isChecked()
        self.config.get(Plugin.Option.SplitChars).value = self._txt_split_char.text()

    def _do_preview(self):
        try:
            result = self._codec.reformat(self._input, self._txt_format.text(), self._txt_split_char.text(),
                                          self._chk_regex.isChecked(), self._chk_new_lines.isChecked())
            self._txt_preview.setPlainText(result)
            return True
        except BaseException as e:
            return False

    def _show_error_message(self, widget, text: str):
        widget.setStyleSheet('QLineEdit { background-color: red }')
        self._error_frame.setHidden(False)
        self._error_text.setText(text)

    def _hide_error_message(self, widgets):
        self._error_frame.setHidden(False)
        self._error_text.setText("")
        for widget in widgets:
            widget.setStyleSheet('QLineEdit {  }')

    def setInput(self, input):
        self._input = input



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import qtawesome
from PyQt5.QtCore import pyqtSignal, Qt, QEvent, QTimer
from PyQt5.QtGui import QPainter
from PyQt5.QtWidgets import QLineEdit


class SearchField(QLineEdit):
    """ Enhances the standard QLineEdit with an icon and additional key-press-events. """

    enterPressed = pyqtSignal()
    escapePressed = pyqtSignal()
    arrowPressed = pyqtSignal()
    closeEvent = pyqtSignal()

    def __init__(self, parent):
        super(__class__, self).__init__(parent)
        self._icon = None
        self._suppress_enter_key = False
        self._closable = False
        self._icon = None
        self._close_icon = None
        self._do_repaint = False
        self.installEventFilter(self)

    def suppressEnterKey(self, suppress=True):
        """ Suppresses the Key_Enter and Key_Return key-press-event when set to True (default = True). """
        self._suppress_enter_key = suppress

    def keyPressEvent(self, evt):
        """ Fires additional signals when a key is pressed. """
        key = evt.key()
        if not self._suppress_enter_key and (key == Qt.Key_Enter or key == Qt.Key_Return):
            self.enterPressed.emit()
        elif key == Qt.Key_Escape:
            self.escapePressed.emit()
        elif key == Qt.Key_Up or key == Qt.Key_Down:
            self.arrowPressed.emit()
        else:
            super(__class__, self).keyPressEvent(evt)

    def eventFilter(self, watched, event):
        """ Catches Mouse-Move- and Mouse-Button-Press-Events for handling Close-Button. """
        if not self._closable:
            return False

        if watched != self:
            return False

        if event.type() != QEvent.MouseMove and event.type() != QEvent.MouseButtonPress:
            return False

        p = event.pos()
        if p.x() > (self.width() - self.height() - 5):
            self.setCursor(Qt.ArrowCursor)
            if event.type() == QEvent.MouseButtonPress:
                self.closeEvent.emit()
        else:
            self.setCursor(Qt.IBeamCursor)

        return False

    def setClosable(self, closable):
        """ When True, this adds a close button to the search field. """
        self._closable = closable
        self._close_icon = qtawesome.icon("fa.times").pixmap(self.height() - 12, self.height() - 12)
        self._do_repaint = True

    def isClosable(self):
        """ Returns whether the search field has a close button. """
        return self._closable

    def flashBackgroundColor(self, color):
        self.setStyleSheet('QLineEdit { background-color: ' + color + ' }')
        QTimer.singleShot(200, self._resetStyleSheet)

    def _resetStyleSheet(self):
        self.setStyleSheet("")

    def setIcon(self, icon):
        """ Sets the specified icon in front of the text. """
        if icon is None:
            self.setTextMargins(1, 1, 1, 1)
        self._icon = icon.pixmap(self.height() - 12, self.height() - 12)
        self._do_repaint = True

    def hasIcon(self):
        return self._icon is not None

    def paintEvent(self, event=None):
        """ Paints the specified icon (if any). """
        super(__class__, self).paintEvent(event)
        if self.hasIcon() or self.isClosable():
            painter = QPainter(self)
            left, top, right, bottom = 1, 1, 1, 1
            if self.hasIcon():
                left = self.height()
                painter.drawPixmap(5, 5, self._icon)
            if self.isClosable():
                right = self.height()
                painter.drawPixmap(self.width() - self.height(), 5, self._close_icon)
            if self._do_repaint:
                self._do_repaint = False
                self.setTextMargins(left, top, right, bottom)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import qtawesome
from PyQt5 import QtCore
from PyQt5.QtCore import pyqtSignal, QRegExp, QPoint, QEvent
from PyQt5.QtGui import QColor, QBrush, QTextCharFormat, QTextCursor, QCursor
from PyQt5.QtWidgets import QFrame, QVBoxLayout, QPlainTextEdit, QAction

from dpp.core import Context
from dpp.ui import SearchField


class PlainView(QFrame):
    """
    Represents the plain-view of the Decoder++ application.
    """

    class EventFilter(QtCore.QObject):

        def __init__(self, parent, context: 'core.context.Context', callback):
            QtCore.QObject.__init__(self, parent)
            self._context = context
            self._callback = callback

        def eventFilter(self, obj, event):
            self._callback(obj, event)
            return QtCore.QObject.eventFilter(self, obj, event)

    def __init__(self, tab_id: str, frame_id: str, text: str, context: Context, parent):
        """
        Initializes the plain view.
        :param text: the text to be shown in the plain-text-edit.
        """
        super(__class__, self).__init__(parent)
        self._context = context
        self._listener = context.listener()
        self._tab_id = tab_id
        self._frame_id = frame_id
        self._logger = logging.getLogger('decoder_plusplus')

        self._plain_text = QPlainTextEdit()
        self.setPlainText(text)
        self._plain_text.setLineWrapMode(QPlainTextEdit.NoWrap)
        self._plain_text.dragEnterEvent = self._on_plain_text_drag_enter_event
        self._plain_text.dropEvent = self._on_plain_text_drop_event
        self._plain_text.textChanged.connect(self._on_plain_text_changed_event)
        self._plain_text.selectionChanged.connect(self._on_plain_text_selection_changed_event)
        self._plain_text.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self._plain_text.customContextMenuRequested.connect(self.showContextMenu)
        self._plain_text.installEventFilter(
            PlainView.EventFilter(self, self._context, self._on_plain_text_focus_changed_event))

        self._listener.textSubmitted.connect(self._on_plain_text_submitted_event)

        self._last_plain_text = self.toPlainText()
        self._last_selected_plain_text = self.toPlainText()

        self._search_field = SearchField(self)
        self._search_field.setClosable(True)
        self._search_field.setIcon(qtawesome.icon("fa.search"))
        self._search_field.setPlaceholderText("Search text")
        self._search_field.escapePressed.connect(self._on_search_field_escape_pressed_event)
        self._search_field.textChanged.connect(self._do_highlight_text)
        self._search_field.closeEvent.connect(self._do_close_search_field)
        self._search_field.setVisible(False)

        # Propagate that a new frame was loaded
        self._context.listener().selectedFrameChanged.emit(self._tab_id, self._frame_id, self.toPlainText())

        layout = QVBoxLayout()
        layout.addWidget(self._plain_text)
        layout.addWidget(self._search_field)
        layout.setContentsMargins(0, 0, 0, 0)

        self.setLayout(layout)

    def id(self) -> str:
        return self._frame_id

    def showContextMenu(self, point: QPoint=None):
        """ Displays a customized context menu for the plain view. """
        if not point:
            point = QCursor.pos()
        context_menu = self._plain_text.createStandardContextMenu()

        textSelected = len(self._plain_text.textCursor().selectedText()) > 0

        open_selection_in_new_tab_action = QAction(self)
        open_selection_in_new_tab_action.setText("Open Selection In New Tab")
        open_selection_in_new_tab_action.setEnabled(textSelected)
        open_selection_in_new_tab_action.triggered.connect(self._on_plain_text_context_menu_open_selection_in_new_tab)
        context_menu.insertAction(context_menu.actions()[0], open_selection_in_new_tab_action)

        separator_action = QAction(self)
        separator_action.setSeparator(True)
        context_menu.insertAction(context_menu.actions()[1], separator_action)

        context_menu.addSeparator()
        wrap_lines_action = QAction(self)
        wrap_lines_action.setText("Wrap Lines")
        wrap_lines_action.setCheckable(True)
        wrap_lines_action.setChecked(self._plain_text.lineWrapMode() == QPlainTextEdit.WidgetWidth)
        wrap_lines_action.triggered.connect(self._on_plain_text_context_menu_wrap_lines)
        context_menu.addAction(wrap_lines_action)
        context_menu.exec(self._plain_text.mapToGlobal(point))

    def _on_plain_text_context_menu_wrap_lines(self, e: QEvent):
        """ Un-/wraps lines when user clicks the wrap-lines action within the plain views context-menu. """
        if self._plain_text.lineWrapMode() == QPlainTextEdit.NoWrap:
            self._plain_text.setLineWrapMode(QPlainTextEdit.WidgetWidth)
        else:
            self._plain_text.setLineWrapMode(QPlainTextEdit.NoWrap)

    def _on_plain_text_context_menu_open_selection_in_new_tab(self, e: QEvent):
        """ Fires a signal that context-menu entry to open selection in new tab was triggered. """
        selectedText = self._plain_text.textCursor().selectedText()
        self._listener.newTabRequested.emit(selectedText)

    def _on_plain_text_drag_enter_event(self, e):
        """ Catches the drag-enter-event which is triggered when media is dragged into the plain text area. """
        data = e.mimeData()
        if data.hasUrls():
            if len(data.urls()) == 1:
                url = data.urls()[0].toString()
                if (url.startswith("file://")):
                    # Accept single file to be dragged into the plain text area.
                    return e.accept()
        elif data.hasText():
            # Accept text to be dragged into the plain text area.
            return e.accept()
        # Ignore everything else.
        return e.ignore()

    def _on_plain_text_drop_event(self, e):
        """ Catches the drop-event which is triggered when media is dropped into the plain text area. """
        data = e.mimeData()
        if data.hasUrls():
            file_path = data.urls()[0].toLocalFile()
            try:
                with open(file_path, mode='rb') as f:
                    # Drops text within (binary) file into plain text area.
                    self._plain_text.setPlainText(f.read().decode('utf-8', errors='surrogateescape'))
            except Exception as e:
                self._logger.error("Error reading file: " + str(e))
        elif data.hasText():
            # Drops text into text field.
            self._plain_text.setPlainText(data.text())

    def _on_plain_text_changed_event(self):
        """ Signals that text has changed and highlights text when search field is active. """
        if self._last_plain_text != self.toPlainText():
            self._last_plain_text = self.toPlainText()
            if self._search_field.isVisible():
                self._do_highlight_text()
            self._context.listener().textChanged.emit(self._tab_id, self._frame_id, self.toPlainText())

    def _on_plain_text_selection_changed_event(self):
        cursor = self._plain_text.textCursor()
        selected_text = cursor.selectedText()
        if selected_text and selected_text != self._last_selected_plain_text:
            self._last_selected_plain_text = selected_text
            self._context.listener().textSelectionChanged.emit(self._tab_id, self._frame_id, selected_text)

    def _on_plain_text_focus_changed_event(self, obj, event):
        if event.type() == QtCore.QEvent.FocusIn and event.reason() == QtCore.Qt.MouseFocusReason:
            if self._plain_text.hasFocus():
                self._context.listener().selectedFrameChanged.emit(self._tab_id, self._frame_id, self.toPlainText())

    def _on_plain_text_submitted_event(self, tab_id: str, frame_id: str, text: str):
        if tab_id == self._tab_id and frame_id == self._frame_id:
            self._plain_text.setPlainText(text)

    def _on_search_field_escape_pressed_event(self):
        """ Closes the search field when search field is focused. """
        if self._search_field.hasFocus() and self._search_field.isVisible():
            self._do_close_search_field()

    def _do_highlight_clear(self):
        """ Clears any highlighting found within the plain text area. """
        self._plain_text.blockSignals(True)
        format = QTextCharFormat()
        format.setForeground(QBrush(QColor("black")))
        cursor = self._plain_text.textCursor()
        cursor.setPosition(0)
        cursor.movePosition(QTextCursor.End, QTextCursor.KeepAnchor, 1)
        cursor.mergeCharFormat(format)
        self._plain_text.blockSignals(False)

    def _do_highlight_text(self):
        """ Highlights text in the plain-view matching the current search-term."""

        def highlight_text(text, format):
            cursor = self._plain_text.textCursor()
            regex = QRegExp(QRegExp.escape(text))

            # Process the displayed document
            pos = 0
            index = regex.indexIn(self._plain_text.toPlainText(), pos)
            while index != -1:
                # Select the matched text and apply the desired format
                cursor.setPosition(index)
                cursor.movePosition(QTextCursor.NextCharacter, QTextCursor.KeepAnchor, len(text))
                cursor.mergeCharFormat(format)
                # Move to the next match
                pos = index + regex.matchedLength()
                index = regex.indexIn(self.toPlainText(), pos)

        self._do_highlight_clear()
        self._plain_text.blockSignals(True)
        searchString = self._search_field.text()
        if searchString:
            format = QTextCharFormat()
            format.setForeground(QBrush(QColor("red")))
            highlight_text(searchString, format)
        self._plain_text.blockSignals(False)

    def _do_open_search_field(self):
        """ Opens the search field. """
        self._search_field.setVisible(True)
        self._do_highlight_text()
        self._search_field.setFocus()

    def _do_close_search_field(self):
        """ Closes the search field. """
        self._do_highlight_clear()
        self._plain_text.setFocus()
        self._search_field.setVisible(False)

    def cutSelectedInputText(self):
        self._plain_text.cut()

    def copySelectedInputText(self):
        self._plain_text.copy()

    def pasteSelectedInputText(self):
        self._plain_text.paste()

    def toggleSearchField(self):
        """ Toggles the search field. """
        if self._search_field.hasFocus() and self._search_field.isVisible():
            self._do_close_search_field()
        else:
            self._do_open_search_field()

    def toPlainText(self):
        """ Returns the plain text of the plain text area. """
        return self._plain_text.toPlainText()

    def setPlainText(self, text):
        """ Sets the text of the text area. """
        # Avoid triggering textChanged-event when setting text manually
        self._plain_text.blockSignals(True)
        self._plain_text.setPlainText(text)
        # Bug: When setting text the cursor position is set to beginning of plain text field.
        # Fix: Manually set cursor position to end of plain text field when setting text.
        self.setCursorPosition(QTextCursor.End)
        self._plain_text.blockSignals(False)

    def setCursorPosition(self, cursor_position: 'QTextCursor.MoveOperation'):
        """ Sets the cursor to the defined position. """
        plain_text_cursor = self._plain_text.textCursor()
        plain_text_cursor.movePosition(cursor_position)
        self._plain_text.setTextCursor(plain_text_cursor)

    def setFocus(self, Qt_FocusReason=None):
        """ Sets the focus to the plain text area. """
        self._plain_text.setFocus()



from dpp.core.plugin import IdentifyPlugin


class Plugin(IdentifyPlugin):
    """
    Detects the file type of the input text based on magic bytes.
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Identify File Type', "Thomas Engel", ["filemagic"], context)

    def _detect_magic_bytes(self, input):
        import magic
        with magic.Magic() as m:
            return m.id_buffer(input)

    def run(self, input):
        return self._detect_magic_bytes(input)


# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from PyQt5.QtCore import QObject, pyqtSignal


class LogFilter(logging.Filter, QObject):
    """ Extends logging.Filter to emit signals for each logged message. """

    logInfoEvent = pyqtSignal('PyQt_PyObject')
    logWarnEvent = pyqtSignal('PyQt_PyObject')
    logErrorEvent = pyqtSignal('PyQt_PyObject')
    logDebugEvent = pyqtSignal('PyQt_PyObject')

    def __init__(self, parent):
        super(logging.Filter, self).__init__()
        super(QObject, self).__init__(parent)

    def filter(self, record):
        if not record.args:
            if record.levelno == logging.INFO:
                self.logInfoEvent.emit(record.msg)
            elif record.levelno == logging.ERROR:
                self.logErrorEvent.emit(record.msg)
            elif record.levelno == logging.WARN:
                self.logWarnEvent.emit(record.msg)
            elif record.levelno == logging.DEBUG:
                self.logDebugEvent.emit(record.msg)
        return True


class LogEntry(object):
    """ Defines a log entry including time, type and message. """

    def __init__(self, time: str, type: str, message: str):
        """
        Initializes a log entry including time, type and message.
        :param time: The time (hour:minute:second) of the log entry (e.g. 00:00:00).
        :param type: The type of the log entry (e.g. INFO, ERROR, ...).
        :param message: The message of the log entry (e.g. 'Hello, world!').
        """
        self._time = time
        self._type = type
        self._message = message

    def time(self) -> str:
        """ Returns the time (hour:minute:second) of the log entry (e.g. 00:00:00). """
        return self._time

    def type(self) -> str:
        """ Returns the type of the log entry (e.g. INFO, ERROR, ...). """
        return self._type

    def message(self) -> str:
        """ Returns the message of the log entry (e.g. 'Hello, world!'). """
        return self._message

    def is_valid(self) -> bool:
        """ Returns whether all log entries are valid (not empty). """
        return bool(self._time) and bool(self._type) and bool(self._message)


from dpp.core.plugin import IdentifyPlugin


class Plugin(IdentifyPlugin):
    """
    Identifies the hash format of the input text based on structure.
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('Identify Hash Format', "Thomas Engel", ["hashid"], context)

    def _detect_hash_format(self, input):
        from hashid import HashID
        return "\n".join(map(str, HashID().identifyHash(input)))

    def run(self, input):
        return self._detect_hash_format(input)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from PyQt5 import QtCore

from dpp.ui.widget.clickable_label import ClickableLabel


class HoverLabel(ClickableLabel):
    """ A label a hover effect. """

    def __init__(self, parent):
        super(HoverLabel, self).__init__(parent)
        self.setHoverEffect(True)

    def setHoverEffect(self, status: bool):
        """ Enables/Disables the hover effect for the IconLabel (default = True). """
        if status:
            self.setStyleSheet("""
            QLabel:hover {
                background-color: rgb(217, 217, 217);
            }

            QLabel[pressed="true"] {
                background-color: rgb(189, 189, 189);     
            }
            """)
        else:
            self.setStyleSheet("")

    def mousePressEvent(self, event):
        """ Updates the stylesheet to indicate that the IconLabel is being pressed. """
        if event.button() == QtCore.Qt.LeftButton:
            self.setProperty("pressed", True)
            self.style().polish(self)
        super(HoverLabel, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        """ Updates the stylesheet to indicate that the IconLabel is being released. """
        if event.button() == QtCore.Qt.LeftButton:
            self.setProperty("pressed", False)
            self.style().polish(self)
        super(HoverLabel, self).mouseReleaseEvent(event)



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import List

import qtawesome
from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import QSortFilterProxyModel, pyqtSignal
from PyQt5.QtGui import QStandardItemModel, QStandardItem
from PyQt5.QtWidgets import QTableView, QHBoxLayout, QToolButton, QVBoxLayout, QFrame, QWidget

from dpp.core.logging import LogEntry
from dpp.ui.widget.dock_widget import DockWidget
from dpp.ui.widget.spacer import VSpacer

class LogDock(DockWidget):
    """ A widget to show log events. """

    clearEvent = pyqtSignal()

    class Filter:

        INFO = "INFO"
        ERROR = "ERROR"
        DEBUG = "DEBUG"

    class ProxyModel(QSortFilterProxyModel):
        def __init__(self):
            super(LogDock.ProxyModel, self).__init__()
            self._filter_info = True
            self._filter_error = True
            self._filter_debug = False

        def filterAcceptsRow(self, rowProc, parentProc):
            sourceModel = self.sourceModel()
            indexProc = sourceModel.index(rowProc, 0, parentProc)
            node = sourceModel.data(indexProc)
            return self._filter("INFO", self._filter_info, node) or \
                   self._filter("ERROR", self._filter_error, node) or \
                   self._filter("DEBUG", self._filter_debug, node)

        def _filter(self, type: str, status: bool, nodeName: str):
            return (nodeName is not None) and (status and type == nodeName)

        def setFilterInfo(self, status: bool):
            self._filter_info = status

        def setFilterError(self, status: bool):
            self._filter_error = status

        def setFilterDebug(self, status: bool):
            self._filter_debug = status

    class Item(QStandardItem):

        def __init__(self, text: str):
            super(LogDock.Item, self).__init__(text)
            self.setEditable(False)

    def __init__(self, log_entries: List[LogEntry], parent: QWidget):
        super(LogDock, self).__init__("Logs", qtawesome.icon("fa.align-left"), parent)
        self._init_button_frame()
        self._init_table_frame(log_entries)
        self.addWidget(self._button_frame)
        self.addWidget(self._table_frame)

    def _init_button_frame(self):
        self._button_frame = QFrame()
        button_layout = QVBoxLayout()

        self._button_filter_info = self._init_tool_button(
            "fa.info-circle", "Filter info messages", True, True, self._filter_info_event)

        self._button_filter_error = self._init_tool_button(
            "fa.exclamation-triangle", "Filter error messages", True, True, self._filter_error_event)

        self._button_filter_debug = self._init_tool_button(
            "fa.bug", "Filter debug messages", True, False, self._filter_debug_event)

        self._button_entries_clear = self._init_tool_button(
            "fa.trash", "Clear log entries", False, False, self._clear_event)

        self._button_filters = {
            LogDock.Filter.INFO: self._button_filter_info,
            LogDock.Filter.ERROR: self._button_filter_error,
            LogDock.Filter.DEBUG: self._button_filter_debug
        }

        button_layout.addWidget(self._button_filter_info)
        button_layout.addWidget(self._button_filter_error)
        button_layout.addWidget(self._button_filter_debug)
        button_layout.addWidget(self._button_entries_clear)
        button_layout.addWidget(VSpacer(self))
        self._button_frame.setLayout(button_layout)

    def _init_tool_button(self, icon_name: str, tool_tip: str, checkable: bool, checked: bool, callback):
        button_filter = QToolButton()
        button_filter.setIcon(qtawesome.icon(icon_name))
        button_filter.setToolTip(tool_tip)
        button_filter.setCheckable(checkable)
        button_filter.setChecked(checked)
        button_filter.clicked.connect(callback)
        return button_filter

    def _init_table_frame(self, log_entries: List[LogEntry]):
        self._table_frame = QFrame()
        table_layout = QHBoxLayout()
        self._table = QTableView()
        self._init_model(log_entries)
        self._init_headers()
        table_layout.addWidget(self._table)
        self._table_frame.setLayout(table_layout)

    def _init_headers(self):
        header = self._table.horizontalHeader()
        header.setDefaultAlignment(QtCore.Qt.AlignLeft)
        header.setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        header.setSectionResizeMode(2, QtWidgets.QHeaderView.Stretch)
        # BUG: Loosing stretch when using ProxyModel
        # FIX: Use setStretchLastSection
        header.setStretchLastSection(True)
        self._table.verticalHeader().hide()

    def _init_model(self, log_entries: List[LogEntry]):
        self._filter_proxy_model = LogDock.ProxyModel()
        model = QStandardItemModel(len(log_entries), 3)
        model.setHorizontalHeaderLabels(["Type", "Time", "Message"])
        for index, log_entry in enumerate(log_entries):
            model.setItem(index, 0, LogDock.Item(log_entry.type()))
            model.setItem(index, 1, LogDock.Item(log_entry.time()))
            model.setItem(index, 2, LogDock.Item(log_entry.message()))
        self._filter_proxy_model.setSourceModel(model)
        self._filter_proxy_model.setFilterKeyColumn(0)
        self._table.setModel(self._filter_proxy_model)

    def addItems(self, *log_entries: List[LogEntry], **kwargs):
        for log_entry in log_entries:
            self.addItem(log_entry)

    def addItem(self, log_entry: LogEntry):
        if log_entry and log_entry.is_valid():
            model = self._table.model().sourceModel()
            nextIndex = model.rowCount()
            model.setItem(nextIndex, 0, LogDock.Item(log_entry.type()))
            model.setItem(nextIndex, 1, LogDock.Item(log_entry.time()))
            model.setItem(nextIndex, 2, LogDock.Item(log_entry.message()))

    def _filter_info_event(self):
        self._filter_proxy_model.setFilterInfo(self._button_filter_info.isChecked())
        self._filter_proxy_model.invalidateFilter()
        self._table.horizontalHeader().setStretchLastSection(True)

    def _filter_error_event(self):
        self._filter_proxy_model.setFilterError(self._button_filter_error.isChecked())
        self._filter_proxy_model.invalidateFilter()
        self._table.horizontalHeader().setStretchLastSection(True)

    def _filter_debug_event(self):
        self._filter_proxy_model.setFilterDebug(self._button_filter_debug.isChecked())
        self._filter_proxy_model.invalidateFilter()
        self._table.horizontalHeader().setStretchLastSection(True)

    def _clear_event(self):
        self._init_headers()
        self._init_model([])
        self.clearEvent.emit()

    def setFilterChecked(self, filter: str, checked: bool):
        if not filter in self._button_filters:
            raise Exception("Unknown filter '{}'.".format(filter))
        if (checked != self._button_filters[filter].isChecked()):
            self._button_filters[filter].click()

    def getFilters(self) -> List[str]:
        return [LogDock.Filter.INFO, LogDock.Filter.ERROR, LogDock.Filter.DEBUG]



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os

import qtawesome
from PyQt5 import QtCore
from PyQt5.QtCore import QSize
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QDialog, QWidget, QHBoxLayout, QTabWidget, QLabel, QFormLayout, QFrame, QVBoxLayout

from dpp.ui import IconLabel, KeyboardShortcutTable, SearchField, HSpacer
from dpp.ui.widget.plugin_tab import PluginTab


class ConfigDialog(QDialog):

    TAB_ABOUT = "About"
    TAB_PLUGINS = "Plugins"
    TAB_KEYBOARD_SHORTCUTS = "Keyboard Shortcuts"

    def __init__(self, parent, context: 'core.context.Context', select_tab_name: str=None):
        """
        Initializes the hidden dialog.
        :param parent: the widget which is calling the dialog.
        :param context: the context of the application.
        :param select_tab_name: the tab name which should be selcted.
        """
        super().__init__(parent)
        self._context = context
        layout = QHBoxLayout()
        tabs = QTabWidget()

        self._about_tab = self._init_about_tab()
        tabs.insertTab(0, self._about_tab, ConfigDialog.TAB_ABOUT)

        self._plugins_tab = self._init_plugins_tab()
        tabs.insertTab(1, self._plugins_tab, ConfigDialog.TAB_PLUGINS)

        self._keyboard_shortcuts_tab = self._init_keyboard_shortcuts_tab()
        tabs.insertTab(2, self._keyboard_shortcuts_tab, ConfigDialog.TAB_KEYBOARD_SHORTCUTS)

        self._init_tab_selection(tabs, select_tab_name)

        layout.addWidget(tabs)

        self.setLayout(layout)
        self.setMinimumWidth(640)
        self.setWindowTitle(" ")

    def _init_tab_selection(self, tabs: QTabWidget, tab_name: str):
        """ Selects the tab with the specified name. """
        for tab_index in range(len(tabs)):
            text = tabs.tabText(tab_index)
            if text == tab_name:
                tabs.setCurrentIndex(tab_index)

    def _init_about_tab(self):
        tab = QWidget(self)
        base_layout = QHBoxLayout()
        base_layout.setAlignment(QtCore.Qt.AlignVCenter)
        logo = os.path.join(self._context.getAppPath(), 'images', 'dpp.png')

        icon_frame = QFrame()
        icon_layout = QHBoxLayout()
        icon_label = IconLabel(self, QIcon(logo))
        icon_label.mouseDoubleClickEvent = lambda evt: self._context.toggleDebugMode()
        icon_label.setFixedSize(QSize(180, 180))
        icon_layout.setAlignment(QtCore.Qt.AlignLeft)
        icon_layout.setContentsMargins(10, 10, 10, 10)
        icon_layout.addWidget(icon_label)
        icon_frame.setLayout(icon_layout)
        base_layout.addWidget(icon_frame)
        form_frame = QFrame(self)
        form_layout = QFormLayout()
        form_layout.setAlignment(QtCore.Qt.AlignLeft)
        form_layout.addRow(QLabel("<font size='18'>Decoder++</font>"), QLabel(""))
        form_layout.addRow(QLabel(""), QLabel(""))
        form_layout.addRow(QLabel("Core-Development: "), QLabel(""))
        form_layout.addRow(QLabel(""), QLabel("Thomas Engel"))
        form_layout.addRow(QLabel("Plugin-Development: "), QLabel(""))
        plugin_authors = self._context.plugins().authors()
        for plugin_author in plugin_authors:
            plugin_author_label = QLabel(plugin_author)
            # Show tooltip with all plugin names.
            # Enable automatic text wrapping of tooltip by using rich-text.
            plugin_author_label.setToolTip("<font>{}</font>".format(", ".join(sorted(set(self._context.plugins().names(author=plugin_author))))))
            form_layout.addRow(QLabel(""), plugin_author_label)
        form_layout.addRow(QLabel("Inspired By: "), QLabel(""))
        form_layout.addRow(QLabel(""), QLabel("PortSwigger's Burp Decoder"))
        form_layout.addRow(QLabel("Powered By: "), QLabel(""))
        form_layout.addRow(QLabel(""), QLabel("PyQt5"))
        form_layout.addRow(QLabel(""), QLabel("QtAwesome"))
        form_layout.addRow(QLabel(""), QLabel("QScintilla"))
        form_layout.addRow(QLabel("Website: "), QLabel(""))
        form_layout.addRow(QLabel(""), QLabel("<a href='https://github.com/bytebutcher/decoder-plus-plus/'>https://github.com/bytebutcher/decoder-plus-plus</a>"))
        form_frame.setLayout(form_layout)
        base_layout.addWidget(form_frame)
        base_layout.addWidget(HSpacer())
        tab.setLayout(base_layout)
        return tab

    def _init_plugins_tab(self):
        return PluginTab(self._context, self)

    def _init_keyboard_shortcuts_tab(self):

        def on_keyboard_shortcut_table_key_press(text):
            # Focus search bar when alpha numeric key was pressed in shortcut table
            shortcut_filter.setFocus()
            shortcut_filter.setText(text)

        frame = QFrame(self)
        frame_layout = QVBoxLayout(self)
        shortcut_filter = SearchField(self)
        shortcut_filter.setPlaceholderText("Search keyboard shortcut...")
        shortcut_filter.setIcon(qtawesome.icon("fa.search"))
        frame_layout.addWidget(shortcut_filter)
        shortcut_table = KeyboardShortcutTable(self, self._context)
        shortcut_table.changed.connect(lambda id, shortcut_key: self._context.updateShortcutKey(id, shortcut_key))
        shortcut_table.keyPressed.connect(on_keyboard_shortcut_table_key_press)
        shortcut_filter.textChanged.connect(shortcut_table.model().setFilterRegExp)
        frame_layout.addWidget(shortcut_table)
        frame_layout.setContentsMargins(20, 20, 20, 20)
        frame.setLayout(frame_layout)
        return frame



import pathlib
import setuptools
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

# This call to setup() does all the work
setup(
    name="decoder-plus-plus",
    version="1.2.1",
    description="An extensible application for penetration testers and software developers to decode/encode data into various formats.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/bytebutcher/decoder-plus-plus",
    author="bytebutcher",
    author_email="thomas.engel.web@gmail.com",
    license="GPL-3.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    packages=setuptools.find_packages(),
    install_requires=[
        'PyQt5==5.15', 
        'QtAwesome==1.0.2',
        'pysha3==1.0.2',
        'passlib==1.7.4',
        'fuzzywuzzy==0.18.0', 
        'filemagic==1.6', 
        'hashid==3.1.4'
    ],
    data_files=[
        ('share/icons/hicolor/scalable/apps', ['data/dpp.png']),
        ('share/applications', ['data/dpp.desktop'])
    ],
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "dpp=dpp.runner:main",
        ]
    },
)



import re

from dpp.core.plugin import DecoderPlugin

class Plugin(DecoderPlugin):
    """
    Decodes a URL-safe BASE64 string.

    URL-safe BASE64 is a BASE64 derivation which remaps following characters:

     Location  Base64  Http64
     --------  ------  ------
       62        +       -
       63        /       _

    Example:

        Input:
            YWJjZGVmZ2hpamtsbW5vcHFyc3R1dnd4eXoKXsKwISLCpyQlJi8oKT0_wrRgPD58ICwuLTs6XyMrJyp-CjAxMjM0NTY3ODk=

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('BASE64 (URL-safe)', "Thomas Engel", ["base64"], context)

    def run(self, text):
        import base64
        text = self._add_missing_padding(text.encode('utf-8', errors="surrogateescape"))
        return base64.urlsafe_b64decode(text).decode('utf-8', errors="surrogateescape")

    def _add_missing_padding(self, text):
        missing_padding = len(text) % 4
        if missing_padding != 0:
            text += b'=' * (4 - missing_padding)
        return text

    def can_decode_input(self, input):
        if len(input) % 4 == 0:
            if input.startswith("0o") and input[2:].isdigit():
                # This looks more like an octal encoding.
                return False
            if not ("-" in input or "_" in input):
                # The urlsafe b64 encoder substitutes - instead of + and _ instead of /.
                # If these characters are not found in the input string it might be b64, but it is not the urlsafe b64.
                return False
            if re.search(r'^(?:[A-Za-z0-9\-_]{4})*(?:[A-Za-z0-9\-_]{2}==|[A-Za-z0-9\-_]{3}=)?$', input):
                return True
        return False



from dpp.core.plugin import DecoderPlugin


class Plugin(DecoderPlugin):
    """
    Decodes an octal string sequence to an ascii string.

    Example:

        Input:
            1411421431441451461471501511521531 \
            5415515615716016116216316416516616 \
            7170171172012136260041042247044045 \
            0460570500510750772641400740761740 \
            4005405605507307213704305304705217 \
            6012060061062063064065066067070071

        Output:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('OCT (str)', "Thomas Engel", [], context)

    def run(self, text):
        return ''.join([chr(int(chunk, 8)) for chunk in self._chunk_string(text, 3)])

    def _chunk_string(self, string, length):
        return [string[0 + i:length + i] for i in range(0, len(string), length)]


# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

class DecoderPlusPlus(object):
    """ The interface of DecoderPlusPlus which is used within the python interactive console. """

    def __init__(self, input: str):
        """
        Initializes DecoderPlusPlus with the specified input.
        :param input: the input which should be transformed.
        """
        self._input = input

    def decode(self) -> 'DecoderPlusPlus':
        """ Returns the decoder interface which encapsulates all possible decoding methods. """
        return Decoder(self._input)

    def encode(self) -> 'DecoderPlusPlus':
        """ Returns the encoder interface which encapsulates all possible encoding methods. """
        return Encoder(self._input)

    def hash(self) -> 'DecoderPlusPlus':
        """ Returns the hash interface which encapsulates all possible hashing methods. """
        return Hasher(self._input)

    def script(self) -> 'DecoderPlusPlus':
        """ Returns the script interface which encapsulates all possible scripting methods. """
        return Script(self._input)

    def identify(self) -> 'DecoderPlusPlus':
        """ Returns the identify interface which encapsulates all possible identify methods. """
        return Identify(self._input)

    def run(self):
        """ Starts the transformation process and returns the transformed input. """
        return self._input


Encoder = type('obj', (DecoderPlusPlus,), {})
Decoder = type('obj', (DecoderPlusPlus,), {})
Hasher = type('obj', (DecoderPlusPlus,), {})
Script = type('obj', (DecoderPlusPlus,), {})
Identify = type('obj', (DecoderPlusPlus,), {})


from dpp.core.plugin import HasherPlugin

class Plugin(HasherPlugin):
    """
    Hashes a string using RIPEMD160.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            1b63bae30eb8be665459c3f2021293811ac2d63b
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('RIPEMD160', "Tim Menapace", ["hashlib"], context)

    def run(self, text):
        import hashlib
        return hashlib.new('ripemd160',text.encode('utf-8', errors='surrogateescape')).hexdigest()


from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QFontMetrics
from PyQt5.QtWidgets import QLabel


class ElidedLabel(QLabel):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.full_text = ""

    def __set_elided_text(self):
        fm = QFontMetrics(self.font())
        super().setText(fm.elidedText(self.full_text, Qt.ElideRight, self.width()))

        self.setToolTip(self.full_text)

    def setText(self, text: str):
        self.full_text = text.split('\n')[0]
        self.__set_elided_text()

    def resizeEvent(self, event) -> None:
        super().resizeEvent(event)
        self.__set_elided_text()

    def minimumSizeHint(self) -> QSize:
        return QSize(0, super().minimumSizeHint().height())




# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
from PyQt5.QtWidgets import QAction


class ActionBuilder:
    """ Builder for QAction. """

    def __init__(self, parent=None):
        self._parent = parent
        self._name = None
        self._enabled = True
        self._callback = None

    def name(self, name: str):
        """ The name of the action (e.g. 'Copy'). """
        self._name = name
        return self

    def enabled(self, status: bool):
        """ The status of the action. When False this action can not be clicked. Default = True. """
        self._enabled = status
        return self

    def callback(self, callback):
        """ The callback which should be triggered when the action is clicked. """
        self._callback = callback # void QAction::triggered(bool checked = false)
        return self

    def build(self) -> QAction:
        """ Builds and returns the QAction with the specified parameters. """
        assert(self._name is not None), "Name is required and should not be None"
        assert(self._callback is not None), "Callback is required and should not be None"
        action = QAction(self._parent)
        action.setText(self._name)
        action.setEnabled(self._enabled)
        action.triggered.connect(self._callback)
        return action



# vim: ts=8:sts=8:sw=8:noexpandtab
#
# This file is part of Decoder++
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import QWidget, QHBoxLayout


class StatusWidget(QWidget):

    DEFAULT = "DEFAULT"
    SUCCESS = "SUCCESS"
    ERROR = "ERROR"

    def __init__(self, parent=None, status="DEFAULT", width=15, height=None):
        super(__class__, self).__init__(parent)
        layout = QHBoxLayout()
        self._widget = QWidget(self)
        if width is not None: self._widget.setFixedWidth(width)
        if height is not None: self._widget.setFixedHeight(height)
        self._status = {
            "DEFAULT": "background-color: rgba( 200, 200, 200, 50% );",
            "SUCCESS": "background-color: rgba( 0, 255, 0, 50% );",
            "ERROR": "background-color: rgba( 255, 0, 0, 50% );"
        }
        self.setStatus(status)
        layout.setContentsMargins(6, 6, 0, 6)

        layout.addWidget(self._widget)
        self.setLayout(layout)

    def setStatus(self, status_name, message=None):
        if status_name not in self._status:
            raise Exception("Unknown status name {}!".format(status_name))
        self._current_status_name = status_name
        self._current_status_message = message
        self._widget.setStyleSheet(self._status[self._current_status_name])
        if message:
            self.setToolTip(message)
        else:
            self.setToolTip("")

    def hasStatus(self, status_name):
        return self._current_status_name == status_name

    def status(self) -> (str, str):
        return self._current_status_name, self._current_status_message



from dpp.core.plugin import EncoderPlugin

class Plugin(EncoderPlugin):
    """
    Replaces a letter with the 13th letter after it.

    Example:

        Input:
            abcdefghijklmnopqrstuvwxyz
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789

        Output:
            nopqrstuvwxyzabcdefghijklm
            ^°!"§$%&/()=?´`<>| ,.-;:_#+'*~
            0123456789
    """

    def __init__(self, context):
        # Name, Author, Dependencies
        super().__init__('ROT13', "Robin Krumnow", ["codecs"], context)

    def run(self, text):
        import codecs
        return codecs.encode(text, 'rot_13')
